"""
Unit tests for ``larray``-compatible ufuncs

Copyright Andrew P. Davison, 2012-2017
"""

from lazyarray import larray, sqrt, cos
import numpy
from numpy.testing import assert_array_equal, assert_array_almost_equal


def test_sqrt_from_array():
    A = larray(numpy.array([1, 4, 9, 16, 25]))
    assert_array_equal(sqrt(A).evaluate(),
                       numpy.arange(1, 6))


def test_sqrt_from_iterator():
    A = larray(iter([1, 4, 9, 16, 25]), shape=(5,))
    assert_array_equal(sqrt(A).evaluate(),
                       numpy.arange(1, 6))


def test_sqrt_from_func():
    A = larray(lambda x: (x + 1) ** 2, shape=(5,))
    assert_array_equal(sqrt(A).evaluate(),
                       numpy.arange(1, 6))


def test_sqrt_normal_array():
    A = numpy.array([1, 4, 9, 16, 25])
    assert_array_equal(sqrt(A),
                       numpy.arange(1, 6))


def test_cos_from_generator():
    def clock():
        for x in numpy.arange(0, 2 * numpy.pi, numpy.pi / 2):
            yield x
    A = larray(clock(), shape=(2, 2))
    assert_array_almost_equal(cos(A).evaluate(),
                              numpy.array([[1.0, 0.0],
                                           [-1.0, 0.0]]),
                              decimal=15)



# encoding: utf-8
"""
lazyarray is a Python package that provides a lazily-evaluated numerical array
class, ``larray``, based on and compatible with NumPy arrays.

Copyright Andrew P. Davison, Joël Chavas and Elodie Legouée (CNRS), 2012-2019
"""

from __future__ import division
import numpy
import operator
from copy import deepcopy
import collections
from functools import wraps
import logging

try:
    from scipy import sparse
    from scipy.sparse import bsr_matrix, coo_matrix, csc_matrix, csr_matrix, dia_matrix, dok_matrix, lil_matrix
    have_scipy = True
except ImportError:
    have_scipy = False


__version__ = "0.3.3"

# stuff for Python 3 compatibility
try:
    long
except NameError:
    long = int

try:
    reduce
except NameError:
    from functools import reduce

try:
    basestring
except NameError:
    basestring = str


logger = logging.getLogger("lazyarray")


def check_shape(meth):
    """
    Decorator for larray magic methods, to ensure that the operand has
    the same shape as the array.
    """
    @wraps(meth)
    def wrapped_meth(self, val):
        if isinstance(val, (larray, numpy.ndarray)) and val.shape:
            if val.shape != self._shape:
                raise ValueError("shape mismatch: objects cannot be broadcast to a single shape")
        return meth(self, val)
    return wrapped_meth


def requires_shape(meth):
    @wraps(meth)
    def wrapped_meth(self, *args, **kwargs):
        if self._shape is None:
            raise ValueError("Shape of larray not specified")
        return meth(self, *args, **kwargs)
    return wrapped_meth


def full_address(addr, full_shape):
    if not (isinstance(addr, numpy.ndarray) and addr.dtype == bool and addr.ndim == len(full_shape)):
        if not isinstance(addr, tuple):
            addr = (addr,)
        if len(addr) < len(full_shape):
            full_addr = [slice(None)] * len(full_shape)
            for i, val in enumerate(addr):
                full_addr[i] = val
            addr = full_addr
    return addr


def partial_shape(addr, full_shape):
    """
    Calculate the size of the sub-array represented by `addr`
    """
    def size(x, max):
        if isinstance(x, (int, long, numpy.integer)):
            return None
        elif isinstance(x, slice):
            y = min(max, x.stop or max)  # slice limits can go past the bounds
            return 1 + (y - (x.start or 0) - 1) // (x.step or 1)
        elif isinstance(x, collections.Sized):
            if hasattr(x, 'dtype') and x.dtype == bool:
                return x.sum()
            else:
                return len(x)
        else:
            raise TypeError("Unsupported index type %s" % type(x))

    addr = full_address(addr, full_shape)
    if isinstance(addr, numpy.ndarray) and addr.dtype == bool:
        return (addr.sum(),)
    elif all(isinstance(x, collections.Sized) for x in addr):
        return (len(addr[0]),)
    else:
        shape = [size(x, max) for (x, max) in zip(addr, full_shape)]
        return tuple([x for x in shape if x is not None])  # remove empty dimensions


def reverse(func):
    """Given a function f(a, b), returns f(b, a)"""
    @wraps(func)
    def reversed_func(a, b):
        return func(b, a)
    reversed_func.__doc__ = "Reversed argument form of %s" % func.__doc__
    reversed_func.__name__ = "reversed %s" % func.__name__
    return reversed_func
# "The hash of a function object is hash(func_code) ^ id(func_globals)" ?
# see http://mail.python.org/pipermail/python-dev/2000-April/003397.html


def lazy_operation(name, reversed=False):
    def op(self, val):
        new_map = deepcopy(self)
        f = getattr(operator, name)
        if reversed:
            f = reverse(f)
        new_map.operations.append((f, val))
        return new_map
    return check_shape(op)


def lazy_inplace_operation(name):
    def op(self, val):
        self.operations.append((getattr(operator, name), val))
        return self
    return check_shape(op)


def lazy_unary_operation(name):
    def op(self):
        new_map = deepcopy(self)
        new_map.operations.append((getattr(operator, name), None))
        return new_map
    return op



class larray(object):
    """
    Optimises storage of and operations on arrays in various ways:
      - stores only a single value if all the values in the array are the same;
      - if the array is created from a function `f(i)` or `f(i,j)`, then
        elements are only evaluated when they are accessed. Any operations
        performed on the array are also queued up to be executed on access.

    Two use cases for the latter are:
      - to save memory for very large arrays by accessing them one row or
        column at a time: the entire array need never be in memory.
      - in parallelized code, different rows or columns may be evaluated
        on different nodes or in different threads.
    """


    def __init__(self, value, shape=None, dtype=None):
        """
        Create a new lazy array.

        `value` : may be an int, long, float, bool, NumPy array, iterator,
                  generator or a function, `f(i)` or `f(i,j)`, depending on the
                  dimensions of the array.

        `f(i,j)` should return a single number when `i` and `j` are integers,
        and a 1D array when either `i` or `j` or both is a NumPy array (in the
        latter case the two arrays must have equal lengths).
        """

        self.dtype = dtype
        self.operations = []
        if isinstance(value, basestring):
            raise TypeError("An larray cannot be created from a string")
        elif isinstance(value, larray):
            if shape is not None and value.shape is not None:
                assert shape == value.shape
            self._shape = shape or value.shape
            self.base_value = value.base_value
            self.dtype = dtype or value.dtype
            self.operations = value.operations  # should deepcopy?

        elif isinstance(value, collections.Sized):  # False for numbers, generators, functions, iterators
            if have_scipy and sparse.issparse(value):  # For sparse matrices
                self.dtype = dtype or value.dtype
            elif not isinstance(value, numpy.ndarray):
                value = numpy.array(value, dtype=dtype)
            elif dtype is not None:
               assert numpy.can_cast(value.dtype, dtype, casting='safe')  # or could convert value to the provided dtype
            if shape and value.shape != shape:
                raise ValueError("Array has shape %s, value has shape %s" % (shape, value.shape))
            self._shape = value.shape
            self.base_value = value

        else:
            assert numpy.isreal(value)  # also True for callables, generators, iterators
            self._shape = shape
            if dtype is None or isinstance(value, dtype):
                self.base_value = value
            else:
                try:
                    self.base_value = dtype(value)
                except TypeError:
                    self.base_value = value

    def __eq__(self, other):
        return self.base_value == other.base_value and self.operations == other.operations and self._shape == other.shape

    def __deepcopy__(self, memo):
        obj = type(self).__new__(type(self))
        if isinstance(self.base_value, VectorizedIterable):  # special case, but perhaps need to rethink
            obj.base_value = self.base_value                 # whether deepcopy is appropriate everywhere
        else:
            try:
                obj.base_value = deepcopy(self.base_value)
            except TypeError:  # base_value cannot be copied, e.g. is a generator (but see generator_tools from PyPI)
                obj.base_value = self.base_value  # so here we create a reference rather than deepcopying - could cause problems
        obj._shape = self._shape
        obj.dtype = self.dtype
        obj.operations = []
        for f, arg in self.operations:
            if isinstance(f, numpy.ufunc):
                obj.operations.append((f, deepcopy(arg)))
            else:
                obj.operations.append((deepcopy(f), deepcopy(arg)))
        return obj

    def __repr__(self):
        return "<larray: base_value=%r shape=%r dtype=%r, operations=%r>" % (self.base_value,
                                                                             self.shape,
                                                                             self.dtype,
                                                                             self.operations)

    def _set_shape(self, value):
        if (hasattr(self.base_value, "shape") and
                self.base_value.shape and   # values of type numpy.float have an empty shape
                    self.base_value.shape != value):
            raise ValueError("Lazy array has fixed shape %s, cannot be changed to %s" % (self.base_value.shape, value))
        self._shape = value
        for op in self.operations:
            if isinstance(op[1], larray):
                op[1].shape = value
    shape = property(fget=lambda self: self._shape,
                     fset=_set_shape, doc="Shape of the array")

    @property
    @requires_shape
    def nrows(self):
        """Size of the first dimension of the array."""
        return self._shape[0]

    @property
    @requires_shape
    def ncols(self):
        """Size of the second dimension (if it exists) of the array."""
        if len(self.shape) > 1:
            return self._shape[1]
        else:
            return 1

    @property
    @requires_shape
    def size(self):
        """Total number of elements in the array."""
        return reduce(operator.mul, self._shape)

    @property
    def is_homogeneous(self):
        """True if all the elements of the array are the same."""
        hom_base = isinstance(self.base_value, (int, long, numpy.integer, float, bool)) \
                   or type(self.base_value) == self.dtype \
                   or (isinstance(self.dtype, type) and isinstance(self.base_value, self.dtype))
        hom_ops = all(obj.is_homogeneous for f, obj in self.operations if isinstance(obj, larray))
        return hom_base and hom_ops

    def _partial_shape(self, addr):
        """
        Calculate the size of the sub-array represented by `addr`
        """
        return partial_shape(addr, self._shape)

    def _homogeneous_array(self, addr):
        self.check_bounds(addr)
        shape = self._partial_shape(addr)
        return numpy.ones(shape, type(self.base_value))

    def _full_address(self, addr):
        return full_address(addr, self._shape)

    def _array_indices(self, addr):
        self.check_bounds(addr)

        def axis_indices(x, max):
            if isinstance(x, (int, long, numpy.integer)):
                return x
            elif isinstance(x, slice):  # need to handle negative values in slice
                return numpy.arange((x.start or 0),
                                    (x.stop or max),
                                    (x.step or 1),
                                    dtype=int)
            elif isinstance(x, collections.Sized):
                if hasattr(x, 'dtype') and x.dtype == bool:
                    return numpy.arange(max)[x]
                else:
                    return numpy.array(x)
            else:
                raise TypeError("Unsupported index type %s" % type(x))
        addr = self._full_address(addr)
        if isinstance(addr, numpy.ndarray) and addr.dtype == bool:
            if addr.ndim == 1:
                return (numpy.arange(self._shape[0])[addr],)
            else:
                raise NotImplementedError()
        elif all(isinstance(x, collections.Sized) for x in addr):
            indices = [numpy.array(x) for x in addr]
            return indices
        else:
            indices = [axis_indices(x, max) for (x, max) in zip(addr, self._shape)]
            if len(indices) == 1:
                return indices
            elif len(indices) == 2:
                if isinstance(indices[0], collections.Sized):
                    if isinstance(indices[1], collections.Sized):
                        mesh_xy = numpy.meshgrid(*indices)
                        return (mesh_xy[0].T, mesh_xy[1].T)  # meshgrid works on (x,y), not (i,j)
                return indices
            else:
                raise NotImplementedError("Only 1D and 2D arrays supported")

    @requires_shape
    def __getitem__(self, addr):
        """
        Return one or more items from the array, as for NumPy arrays.

        `addr` may be a single integer, a slice, a NumPy boolean array or a
        NumPy integer array.
        """
        return self._partially_evaluate(addr, simplify=False)

    def _partially_evaluate(self, addr, simplify=False):
        """
        Return part of the lazy array.
        """
        if self.is_homogeneous:
            if simplify:
                base_val = self.base_value
            else:
                base_val = self._homogeneous_array(addr) * self.base_value
        elif isinstance(self.base_value, (int, long, numpy.integer, float, bool)):
            base_val = self._homogeneous_array(addr) * self.base_value
        elif isinstance(self.base_value, numpy.ndarray):
            base_val = self.base_value[addr]
        elif have_scipy and sparse.issparse(self.base_value):  # For sparse matrices larr[2, :]
            base_val = self.base_value[addr]
        elif callable(self.base_value):
            indices = self._array_indices(addr)
            base_val = self.base_value(*indices)
            if isinstance(base_val, numpy.ndarray) and base_val.shape == (1,):
                base_val = base_val[0]
        elif hasattr(self.base_value, "lazily_evaluate"):
            base_val = self.base_value.lazily_evaluate(addr, shape=self._shape)
        elif isinstance(self.base_value, VectorizedIterable):
            partial_shape = self._partial_shape(addr)
            if partial_shape:
                n = reduce(operator.mul, partial_shape)
            else:
                n = 1
            base_val = self.base_value.next(n)  # note that the array contents will depend on the order of access to elements
            if n == 1:
                base_val = base_val[0]
            elif partial_shape and base_val.shape != partial_shape:
                base_val = base_val.reshape(partial_shape)
        elif isinstance(self.base_value, collections.Iterator):
            raise NotImplementedError("coming soon...")
        else:
            raise ValueError("invalid base value for array (%s)" % self.base_value)
        return self._apply_operations(base_val, addr, simplify=simplify)

    @requires_shape
    def check_bounds(self, addr):
        """
        Check whether the given address is within the array bounds.
        """
        def is_boolean_array(arr):
            return hasattr(arr, 'dtype') and arr.dtype == bool

        def check_axis(x, size):
            if isinstance(x, (int, long, numpy.integer)):
                lower = upper = x
            elif isinstance(x, slice):
                lower = x.start or 0
                upper = min(x.stop or size - 1, size - 1)  # slices are allowed to go past the bounds
            elif isinstance(x, collections.Sized):
                if is_boolean_array(x):
                    lower = 0
                    upper = x.size - 1
                else:
                    if len(x) == 0:
                        raise ValueError("Empty address component (address was %s)" % str(addr))
                    if hasattr(x, "min"):
                        lower = x.min()
                    else:
                        lower = min(x)
                    if hasattr(x, "max"):
                        upper = x.max()
                    else:
                        upper = max(x)
            else:
                raise TypeError("Invalid array address: %s (element of type %s)" % (str(addr), type(x)))
            if (lower < -size) or (upper >= size):
                raise IndexError("Index out of bounds")
        full_addr = self._full_address(addr)
        if isinstance(addr, numpy.ndarray) and addr.dtype == bool:
            if len(addr.shape) > len(self._shape):
                raise IndexError("Too many indices for array")
            for xmax, size in zip(addr.shape, self._shape):
                upper = xmax - 1
                if upper >= size:
                    raise IndexError("Index out of bounds")
        else:
            for i, size in zip(full_addr, self._shape):
                check_axis(i, size)

    def apply(self, f):
        """
        Add the function `f(x)` to the list of the operations to be performed,
        where `x` will be a scalar or a numpy array.

        >>> m = larray(4, shape=(2,2))
        >>> m.apply(numpy.sqrt)
        >>> m.evaluate()
        array([[ 2.,  2.],
               [ 2.,  2.]])
        """
        self.operations.append((f, None))

    def _apply_operations(self, x, addr=None, simplify=False):
        for f, arg in self.operations:
            if arg is None:
                x = f(x)
            elif isinstance(arg, larray):
                if addr is None:
                    x = f(x, arg.evaluate(simplify=simplify))
                else:
                    x = f(x, arg._partially_evaluate(addr, simplify=simplify))

            else:
                x = f(x, arg)
        return x

    @requires_shape
    def evaluate(self, simplify=False, empty_val=0):
        """
        Return the lazy array as a real NumPy array.

        If the array is homogeneous and ``simplify`` is ``True``, return a
        single numerical value.
        """
        # need to catch the situation where a generator-based larray is evaluated a second time
        if self.is_homogeneous:
            if simplify:
                x = self.base_value
            else:
                x = self.base_value * numpy.ones(self._shape, dtype=self.dtype)
        elif isinstance(self.base_value, (int, long, numpy.integer, float, bool, numpy.bool_)):
            x = self.base_value * numpy.ones(self._shape, dtype=self.dtype)
        elif isinstance(self.base_value, numpy.ndarray):
            x = self.base_value
        elif callable(self.base_value):
            x = numpy.array(numpy.fromfunction(self.base_value, shape=self._shape, dtype=int), dtype=self.dtype)
        elif hasattr(self.base_value, "lazily_evaluate"):
            x = self.base_value.lazily_evaluate(shape=self._shape)
        elif isinstance(self.base_value, VectorizedIterable):
            x = self.base_value.next(self.size)
            if x.shape != self._shape:
                x = x.reshape(self._shape)
        elif have_scipy and sparse.issparse(self.base_value):  # For sparse matrices
            if empty_val!=0:
                x = self.base_value.toarray((sparse.csc_matrix))
                x = numpy.where(x, x, numpy.nan)
            else:
                x = self.base_value.toarray((sparse.csc_matrix))
        elif isinstance(self.base_value, collections.Iterator):
            x = numpy.fromiter(self.base_value, dtype=self.dtype or float, count=self.size)
            if x.shape != self._shape:
                x = x.reshape(self._shape)
        else:
            raise ValueError("invalid base value for array")
        return self._apply_operations(x, simplify=simplify)

    def __call__(self, arg):
        if callable(self.base_value):
            if isinstance(arg, larray):
                new_map = deepcopy(arg)
            elif callable(arg):
                new_map = larray(arg)
            else:
                raise Exception("Argument must be either callable or an larray.")
            new_map.operations.append((self.base_value, None))
            new_map.operations.extend(self.operations)
            return new_map
        else:
            raise Exception("larray is not callable")

    __iadd__ = lazy_inplace_operation('add')
    __isub__ = lazy_inplace_operation('sub')
    __imul__ = lazy_inplace_operation('mul')
    __idiv__ = lazy_inplace_operation('div')
    __ipow__ = lazy_inplace_operation('pow')

    __add__ = lazy_operation('add')
    __radd__ = __add__
    __sub__ = lazy_operation('sub')
    __rsub__ = lazy_operation('sub', reversed=True)
    __mul__ = lazy_operation('mul')
    __rmul__ = __mul__
    __div__ = lazy_operation('div')
    __rdiv__ = lazy_operation('div', reversed=True)
    __truediv__ = lazy_operation('truediv')
    __rtruediv__ = lazy_operation('truediv', reversed=True)
    __pow__ = lazy_operation('pow')

    __lt__ = lazy_operation('lt')
    __gt__ = lazy_operation('gt')
    __le__ = lazy_operation('le')
    __ge__ = lazy_operation('ge')

    __neg__ = lazy_unary_operation('neg')
    __pos__ = lazy_unary_operation('pos')
    __abs__ = lazy_unary_operation('abs')


def _build_ufunc(func):
    """Return a ufunc that works with lazy arrays"""
    def larray_compatible_ufunc(x):
        if isinstance(x, larray):
            y = deepcopy(x)
            y.apply(func)
            return y
        else:
            return func(x)
    return larray_compatible_ufunc


class VectorizedIterable(object):
    """
    Base class for any class which has a method `next(n)`, i.e., where you
    can choose how many values to return rather than just returning one at a
    time.
    """
    pass


# build lazy-array compatible versions of NumPy ufuncs
namespace = globals()
for name in dir(numpy):
    obj = getattr(numpy, name)
    if isinstance(obj, numpy.ufunc):
        namespace[name] = _build_ufunc(obj)






import numpy as np
from numpy.testing import assert_array_equal
from lazyarray import larray
from timeit import repeat


def test_function(i, j):
    return i * i + 2 * i * j + 3


def array_from_function_full(f, shape):
    return np.fromfunction(f, shape)


def larray_from_function_full(f, shape):
    return larray(f, shape).evaluate()


def array_from_function_slice(f, shape):
    return np.fromfunction(f, shape)[:, 0:-1:10]


def larray_from_function_slice(f, shape):
    return larray(f, shape)[:, 0:shape[1] - 1:10]


if __name__ == "__main__":
    assert_array_equal(array_from_function_full(test_function, (5000, 5000)),
                       larray_from_function_full(test_function, (5000, 5000)))

    print "Array from function: full array"
    print(repeat('array_from_function_full(test_function, (5000, 5000))',
                 setup='from __main__ import array_from_function_full, test_function',
                 number=1, repeat=5))

    print "Lazy array from function: full array"
    print(repeat('larray_from_function_full(test_function, (5000, 5000))',
                 setup='from __main__ import larray_from_function_full, test_function',
                 number=1, repeat=5))

    assert_array_equal(array_from_function_slice(test_function, (5000, 5000)),
                       larray_from_function_slice(test_function, (5000, 5000)))
    print "Array from function: slice"
    print(repeat('array_from_function_slice(test_function, (5000, 5000))',
                 setup='from __main__ import array_from_function_slice, test_function',
                 number=1, repeat=5))

    print "Lazy array from function: slice"
    print(repeat('larray_from_function_slice(test_function, (5000, 5000))',
                 setup='from __main__ import larray_from_function_slice, test_function',
                 number=1, repeat=5))



# Support creating lazy arrays from SciPy sparse matrices 
#
# 1 program for the 7 sparse matrices classes : 
#
# csc_matrix(arg1[, shape, dtype, copy])	            Compressed Sparse Column matrix
# csr_matrix(arg1[, shape, dtype, copy])	            Compressed Sparse Row matrix
# bsr_matrix(arg1[, shape, dtype, copy, blocksize])	    Block Sparse Row matrix
# lil_matrix(arg1[, shape, dtype, copy])	            Row-based linked list sparse matrix
# dok_matrix(arg1[, shape, dtype, copy])	            Dictionary Of Keys based sparse matrix.
# coo_matrix(arg1[, shape, dtype, copy])	            A sparse matrix in COOrdinate format.
# dia_matrix(arg1[, shape, dtype, copy])	            Sparse matrix with DIAgonal storage
#


import numpy as np
from lazyarray import larray
from scipy import sparse 
import random


################
# Random numbers
################
i = random.randint(-100, 100)
j = random.randint(-100, 100)
k = random.randint(-100, 100)
l = random.randint(-100, 100)
m = random.randint(-100, 100)
n = random.randint(-100, 100)
p = random.randint(-100, 100)
q = random.randint(-100, 100)
r = random.randint(-100, 100)

################
# An example
################
#i = 1
#j = 2
#k = 0
#l = 0
#m = 0
#n = 3
#p = 1
#q = 0
#r = 4

#print "i =", i
#print "j =", j
#print "k =", k
#print "l =", l
#print "m =", m
#print "n =", n
#print "p =", p
#print "q  =", q
#print "r =", r    


##############################################################
# Definition of an array
##############################################################

def test_function_array_general():  
    A = np.array([[i, j, k], [l, m, n], [p, q, r]])
    #print "A ="
    #print A
    return A


##############################################################
# Definition of 7 sparse matrices
##############################################################

def sparse_csc_matrices():
    csc = sparse.csc_matrix([[i, j, k], [l, m, n], [p, q, r]])
    #print "csc matrices ="
    #print csc
    return csc

def sparse_csr_matrices():
    csr = sparse.csr_matrix([[i, j, k], [l, m, n], [p, q, r]])
    #print "csr matrices ="
    #print csr
    return csr

def sparse_bsr_matrices():
    bsr = sparse.bsr_matrix([[i, j, k], [l, m, n], [p, q, r]])
    #print "bsr matrices ="
    #print bsr
    return bsr

def sparse_lil_matrices():
    lil = sparse.lil_matrix([[i, j, k], [l, m, n], [p, q, r]])
    #print "lil matrices ="
    #print lil
    return lil

def sparse_dok_matrices():
    dok = sparse.dok_matrix([[i, j, k], [l, m, n], [p, q, r]])
    #print "dok matrices ="
    #print dok
    return dok

def sparse_coo_matrices():
    coo = sparse.coo_matrix([[i, j, k], [l, m, n], [p, q, r]])
    #print "coo matrices ="
    #print coo
    return coo

def sparse_dia_matrices():
    dia = sparse.dia_matrix([[i, j, k], [l, m, n], [p, q, r]])
    #print "dia matrices ="
    #print dia
    return dia

    

if __name__ == "__main__": 


##############################################################
# Call test_function_array_general
# Create a sparse matrix from array
# There are 7 sparse matrices
##############################################################

    #print "Array general ="
    test_function_array_general()    
    #print "Array ="
    #print test_function_array_general()
    
#    print "----"

#    print "Sparse array csc general ="
    sA_csc_general = sparse.csc_matrix(test_function_array_general())
    #print ("sparse csc matrices", sparse.csc_matrix(test_function_array_general()))
    #print "sparse csc matrices ="
    #print sA_csc_general
#    print "----"
#    print "Sparse array csr general ="
    sA_csr = sparse.csr_matrix(test_function_array_general())
    #print ("sparse csr matrices", sparse.csr_matrix(test_function_array_general()))
    #print "sparse csr matrices ="
    #print sA_csr
#    print "----"
#    print "Sparse array bsr general  ="
    sA_bsr = sparse.bsr_matrix(test_function_array_general())
#    print ("sparse bsr matrices", sparse.bsr_matrix(test_function_array_general()))
#    print "sparse bsr matrices ="
#    print sA_bsr
#    print "----"
#    print "Sparse array lil general ="
    sA_lil = sparse.lil_matrix(test_function_array_general())
#    print ("sparse lil matrices", sparse.lil_matrix(test_function_array_general()))
#    print "sparse lil matrices ="
#    print sA_lil
#    print "----"
#    print "Sparse array dok general ="
    sA_dok = sparse.dok_matrix(test_function_array_general())
#    print ("sparse dok matrices", sparse.dok_matrix(test_function_array_general()))
#    print "sparse dok matrices ="
#    print sA_dok
#    print "----"
#    print "Sparse array coo general ="
    sA_coo = sparse.coo_matrix(test_function_array_general())
#    print ("sparse coo matrices", sparse.coo_matrix(test_function_array_general()))
#    print "sparse coo matrices ="
#    print sA_coo
#    print "----"
#    print "Sparse array dia general ="
    sA_dia = sparse.dia_matrix(test_function_array_general())
#    print ("sparse dia matrices", sparse.dia_matrix(test_function_array_general()))
#    print "sparse dia matrices ="
#    print sA_dia


#print "----------------------------------------------------------------------"
 

 ##############################################################
 # Call the sparse matrices
 # Create a lazy array from sparse matrices
 ##############################################################
 
 
Array_csc_matrices = sparse_csc_matrices().toarray()
#print "Array csc matrices ="
#print Array_csc_matrices
  
Array_csr_matrices = sparse_csr_matrices().toarray()
#print "Array csr matrices ="
#print Array_csr_matrices

Array_bsr_matrices = sparse_bsr_matrices().toarray()
#print "Array bsr matrices ="
#print Array_bsr_matrices

Array_lil_matrices = sparse_lil_matrices().toarray()
#print "Array lil matrices ="
#print Array_lil_matrices
    
Array_dok_matrices = sparse_dok_matrices().toarray()
#print "Array dok matrices ="
#print Array_dok_matrices

Array_coo_matrices = sparse_coo_matrices().toarray()
#print "Array coo matrices ="
#print Array_coo_matrices

Array_dia_matrices = sparse_dia_matrices().toarray()
#print "Array dia matrices ="
#print Array_dia_matrices


# -*- coding: utf-8 -*-

from distutils.core import setup

setup(
    name='lazyarray',
    version='0.3.3',
    py_modules=['lazyarray'],
    license='Modified BSD',
    author="Andrew P. Davison",
    author_email="andrew.davison@unic.cnrs-gif.fr",
    url="http://github.com/NeuralEnsemble/lazyarray/",
    description="a Python package that provides a lazily-evaluated numerical array class, larray, based on and compatible with NumPy arrays.",
    long_description=open('README.rst').read(),
    install_requires=[
        "numpy >= 1.8",
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
    ]
)



# encoding: utf-8
"""
Unit tests for ``larray`` class

Copyright Andrew P. Davison, Joël Chavas and Elodie Legouée (CNRS), 2012-2017
"""

from lazyarray import larray, VectorizedIterable, sqrt, partial_shape
import numpy
from nose.tools import assert_raises, assert_equal
from nose import SkipTest
from numpy.testing import assert_array_equal, assert_array_almost_equal
import operator
from copy import deepcopy
from scipy.sparse import bsr_matrix, coo_matrix, csc_matrix, csr_matrix, dia_matrix, dok_matrix, lil_matrix



class MockRNG(VectorizedIterable):

    def __init__(self, start, delta):
        self.start = start
        self.delta = delta

    def next(self, n):
        s = self.start
        self.start += n * self.delta
        return s + self.delta * numpy.arange(n)


# test larray
def test_create_with_int():
    A = larray(3, shape=(5,))
    assert A.shape == (5,)
    assert A.evaluate(simplify=True) == 3


def test_create_with_int_and_dtype():
    A = larray(3, shape=(5,), dtype=float)
    assert A.shape == (5,)
    assert A.evaluate(simplify=True) == 3


def test_create_with_float():
    A = larray(3.0, shape=(5,))
    assert A.shape == (5,)
    assert A.evaluate(simplify=True) == 3.0


def test_create_with_list():
    A = larray([1, 2, 3], shape=(3,))
    assert A.shape == (3,)
    assert_array_equal(A.evaluate(), numpy.array([1, 2, 3]))


def test_create_with_array():
    A = larray(numpy.array([1, 2, 3]), shape=(3,))
    assert A.shape == (3,)
    assert_array_equal(A.evaluate(), numpy.array([1, 2, 3]))


def test_create_with_array_and_dtype():
    A = larray(numpy.array([1, 2, 3]), shape=(3,), dtype=int)
    assert A.shape == (3,)
    assert_array_equal(A.evaluate(), numpy.array([1, 2, 3]))


def test_create_with_generator():
    def plusone():
        i = 0
        while True:
            yield i
            i += 1
    A = larray(plusone(), shape=(5, 11))
    assert_array_equal(A.evaluate(),
                       numpy.arange(55).reshape((5, 11)))


def test_create_with_function1D():
    A = larray(lambda i: 99 - i, shape=(3,))
    assert_array_equal(A.evaluate(),
                       numpy.array([99, 98, 97]))


def test_create_with_function1D_and_dtype():
    A = larray(lambda i: 99 - i, shape=(3,), dtype=float)
    assert_array_equal(A.evaluate(),
                       numpy.array([99.0, 98.0, 97.0]))


def test_create_with_function2D():
    A = larray(lambda i, j: 3 * j - 2 * i, shape=(2, 3))
    assert_array_equal(A.evaluate(),
                       numpy.array([[0, 3, 6],
                                    [-2, 1, 4]]))


def test_create_inconsistent():
    assert_raises(ValueError, larray, [1, 2, 3], shape=4)


def test_create_with_string():
    assert_raises(TypeError, larray, "123", shape=3)


def test_create_with_larray():
    A = 3 + larray(lambda i: 99 - i, shape=(3,))
    B = larray(A, shape=(3,), dtype=int)
    assert_array_equal(B.evaluate(),
                       numpy.array([102, 101, 100]))


## For sparse matrices
def test_create_with_sparse_array():
    row = numpy.array([0, 2, 2, 0, 1, 2])
    col = numpy.array([0, 0, 1, 2, 2, 2])
    data = numpy.array([1, 2, 3, 4, 5, 6])
    bsr = larray(bsr_matrix((data, (row, col)), shape=(3, 3))) # For bsr_matrix
    coo = larray(coo_matrix((data, (row, col)), shape=(3, 3))) # For coo_matrix
    csc = larray(csc_matrix((data, (row, col)), shape=(3, 3))) # For csc_matrix
    csr = larray(csr_matrix((data, (row, col)), shape=(3, 3))) # For csr_matrix
    data_dia = numpy.array([[1, 2, 3, 4]]).repeat(3, axis=0) # For dia_matrix
    offsets_dia = numpy.array([0, -1, 2]) # For dia_matrix
    dia = larray(dia_matrix((data_dia, offsets_dia), shape=(4, 4))) # For dia_matrix
    dok = larray(dok_matrix(((row, col)), shape=(3, 3))) # For dok_matrix
    lil = larray(lil_matrix(data, shape=(3, 3))) # For lil_matrix  
    assert bsr.shape == (3, 3)
    assert coo.shape == (3, 3)
    assert csc.shape == (3, 3)
    assert csr.shape == (3, 3)
    assert dia.shape == (4, 4)
    assert dok.shape == (2, 6)
    assert lil.shape == (1, 6)

    def test_evaluate_with_sparse_array():
        assert_array_equal(bsr.evaluate(), bsr_matrix((data, (row, col))).toarray()) # For bsr_matrix
        assert_array_equal(coo.evaluate(), coo_matrix((data, (row, col))).toarray()) # For coo_matrix
        assert_array_equal(csc.evaluate(), csc_matrix((data, (row, col))).toarray()) # For csc_matrix
        assert_array_equal(csr.evaluate(), csr_matrix((data, (row, col))).toarray()) # For csr_matrix
        assert_array_equal(dia.evaluate(), dia_matrix((data_dia, (row, col))).toarray()) # For dia_matrix
        assert_array_equal(dok.evaluate(), dok_matrix((data, (row, col))).toarray()) # For dok_matrix
        assert_array_equal(lil.evaluate(), lil_matrix((data, (row, col))).toarray()) # For lil_matrix

    def test_multiple_operations_with_sparse_array():
        # For bsr_matrix
        bsr0 = bsr /100.0
        bsr1 = 0.2 + bsr0
        assert_array_almost_equal(bsr0.evaluate(), np.array([[0.01, 0., 0.04], [0., 0., 0.05], [0.02, 0.03, 0.06]]))
        assert_array_almost_equal(bsr0.evaluate(), np.array([[0.21, 0.2, 0.24], [0.2, 0.2, 0.25], [0.22, 0.23, 0.26]]))
        # For coo_matrix
        coo0 = coo /100.0
        coo1 = 0.2 + coo0
        assert_array_almost_equal(coo0.evaluate(), np.array([[0.01, 0., 0.04], [0., 0., 0.05], [0.02, 0.03, 0.06]]))
        assert_array_almost_equal(coo0.evaluate(), np.array([[0.21, 0.2, 0.24], [0.2, 0.2, 0.25], [0.22, 0.23, 0.26]]))
        # For csc_matrix
        csc0 = csc /100.0
        csc1 = 0.2 + csc0
        assert_array_almost_equal(csc0.evaluate(), np.array([[0.01, 0., 0.04], [0., 0., 0.05], [0.02, 0.03, 0.06]]))
        assert_array_almost_equal(csc0.evaluate(), np.array([[0.21, 0.2, 0.24], [0.2, 0.2, 0.25], [0.22, 0.23, 0.26]]))
        # For csr_matrix
        csr0 = csr /100.0
        csr1 = 0.2 + csr0
        assert_array_almost_equal(csc0.evaluate(), np.array([[0.01, 0., 0.04], [0., 0., 0.05], [0.02, 0.03, 0.06]]))
        assert_array_almost_equal(csc0.evaluate(), np.array([[0.21, 0.2, 0.24], [0.2, 0.2, 0.25], [0.22, 0.23, 0.26]]))    
        # For dia_matrix
        dia0 = dia /100.0
        dia1 = 0.2 + dia0
        assert_array_almost_equal(dia0.evaluate(), np.array([[0.01, 0.02, 0.03, 0.04]]))
        assert_array_almost_equal(dia1.evaluate(), np.array([[0.21, 0.22, 0.23, 0.24]]))
         # For dok_matrix
        dok0 = dok /100.0
        dok1 = 0.2 + dok0
        assert_array_almost_equal(dok0.evaluate(), np.array([[0., 0.02, 0.02, 0., 0.01, 0.02], [0., 0., 0.01, 0.02, 0.02, 0.02]]))
        assert_array_almost_equal(dok1.evaluate(), np.array([[0.2, 0.22, 0.22, 0.2, 0.21, 0.22], [0.2, 0.2, 0.21, 0.22, 0.22, 0.22]]))
         # For lil_matrix
        lil0 = lil /100.0
        lil1 = 0.2 + lil0
        assert_array_almost_equal(lil0.evaluate(), np.array([[0.01, 0.02, 0.03, 0.04, 0.05, 0.06]]))
        assert_array_almost_equal(lil1.evaluate(), np.array([[0.21, 0.22, 0.23, 0.24, 0.25, 0.26]]))


    def test_getitem_from_2D_sparse_array():
        assert_raises(IndexError, bsr.__getitem__, (3, 0))
        assert_raises(IndexError, coo.__getitem__, (3, 0))
        assert_raises(IndexError, csc.__getitem__, (3, 0))
        assert_raises(IndexError, csr.__getitem__, (3, 0))
        assert_raises(IndexError, dia.__getitem__, (3, 0))
        assert_raises(IndexError, dok.__getitem__, (3, 0))
        assert_raises(IndexError, lil.__getitem__, (3, 0))
    

# def test_columnwise_iteration_with_flat_array():
# m = larray(5, shape=(4,3)) # 4 rows, 3 columns
#    cols = [col for col in m.by_column()]
#    assert_equal(cols, [5, 5, 5])
#
# def test_columnwise_iteration_with_structured_array():
#    input = numpy.arange(12).reshape((4,3))
# m = larray(input, shape=(4,3)) # 4 rows, 3 columns
#    cols = [col for col in m.by_column()]
#    assert_array_equal(cols[0], input[:,0])
#    assert_array_equal(cols[2], input[:,2])
#
# def test_columnwise_iteration_with_function():
#    input = lambda i,j: 2*i + j
#    m = larray(input, shape=(4,3))
#    cols = [col for col in m.by_column()]
#    assert_array_equal(cols[0], numpy.array([0, 2, 4, 6]))
#    assert_array_equal(cols[1], numpy.array([1, 3, 5, 7]))
#    assert_array_equal(cols[2], numpy.array([2, 4, 6, 8]))
#
# def test_columnwise_iteration_with_flat_array_and_mask():
# m = larray(5, shape=(4,3)) # 4 rows, 3 columns
#    mask = numpy.array([True, False, True])
#    cols = [col for col in m.by_column(mask=mask)]
#    assert_equal(cols, [5, 5])
#
# def test_columnwise_iteration_with_structured_array_and_mask():
#    input = numpy.arange(12).reshape((4,3))
# m = larray(input, shape=(4,3)) # 4 rows, 3 columns
#    mask = numpy.array([False, True, True])
#    cols = [col for col in m.by_column(mask=mask)]
#    assert_array_equal(cols[0], input[:,1])
#    assert_array_equal(cols[1], input[:,2])


def test_size_related_properties():
    m1 = larray(1, shape=(9, 7))
    m2 = larray(1, shape=(13,))
    m3 = larray(1)
    assert_equal(m1.nrows, 9)
    assert_equal(m1.ncols, 7)
    assert_equal(m1.size, 63)
    assert_equal(m2.nrows, 13)
    assert_equal(m2.ncols, 1)
    assert_equal(m2.size, 13)
    assert_raises(ValueError, lambda: m3.nrows)
    assert_raises(ValueError, lambda: m3.ncols)
    assert_raises(ValueError, lambda: m3.size)


def test_evaluate_with_flat_array():
    m = larray(5, shape=(4, 3))
    assert_array_equal(m.evaluate(), 5 * numpy.ones((4, 3)))


def test_evaluate_with_structured_array():
    input = numpy.arange(12).reshape((4, 3))
    m = larray(input, shape=(4, 3))
    assert_array_equal(m.evaluate(), input)


def test_evaluate_with_functional_array():
    input = lambda i, j: 2 * i + j
    m = larray(input, shape=(4, 3))
    assert_array_equal(m.evaluate(),
                       numpy.array([[0, 1, 2],
                                    [2, 3, 4],
                                    [4, 5, 6],
                                    [6, 7, 8]]))


def test_evaluate_with_vectorized_iterable():
    input = MockRNG(0, 1)
    m = larray(input, shape=(7, 3))
    assert_array_equal(m.evaluate(),
                       numpy.arange(21).reshape((7, 3)))


def test_evaluate_twice_with_vectorized_iterable():
    input = MockRNG(0, 1)
    m1 = larray(input, shape=(7, 3)) + 3
    m2 = larray(input, shape=(7, 3)) + 17
    assert_array_equal(m1.evaluate(),
                       numpy.arange(3, 24).reshape((7, 3)))
    assert_array_equal(m2.evaluate(),
                       numpy.arange(38, 59).reshape((7, 3)))


def test_evaluate_structured_array_size_1_simplify():
    m = larray([5.0], shape=(1,))
    assert_equal(m.evaluate(simplify=True), 5.0)
    n = larray([2.0], shape=(1,))
    assert_equal((m/n).evaluate(simplify=True), 2.5)


def test_iadd_with_flat_array():
    m = larray(5, shape=(4, 3))
    m += 2
    assert_array_equal(m.evaluate(), 7 * numpy.ones((4, 3)))
    assert_equal(m.base_value, 5)
    assert_equal(m.evaluate(simplify=True), 7)


def test_add_with_flat_array():
    m0 = larray(5, shape=(4, 3))
    m1 = m0 + 2
    assert_equal(m1.evaluate(simplify=True), 7)
    assert_equal(m0.evaluate(simplify=True), 5)


def test_lt_with_flat_array():
    m0 = larray(5, shape=(4, 3))
    m1 = m0 < 10
    assert_equal(m1.evaluate(simplify=True), True)
    assert_equal(m0.evaluate(simplify=True), 5)


def test_lt_with_structured_array():
    input = numpy.arange(12).reshape((4, 3))
    m0 = larray(input, shape=(4, 3))
    m1 = m0 < 5
    assert_array_equal(m1.evaluate(simplify=True), input < 5)


def test_structured_array_lt_array():
    input = numpy.arange(12).reshape((4, 3))
    m0 = larray(input, shape=(4, 3))
    comparison = 5 * numpy.ones((4, 3))
    m1 = m0 < comparison
    assert_array_equal(m1.evaluate(simplify=True), input < comparison)


def test_rsub_with_structured_array():
    m = larray(numpy.arange(12).reshape((4, 3)))
    assert_array_equal((11 - m).evaluate(),
                       numpy.arange(11, -1, -1).reshape((4, 3)))


def test_inplace_mul_with_structured_array():
    m = larray((3 * x for x in range(4)), shape=(4,))
    m *= 7
    assert_array_equal(m.evaluate(),
                       numpy.arange(0, 84, 21))


def test_abs_with_structured_array():
    m = larray(lambda i, j: i - j, shape=(3, 4))
    assert_array_equal(abs(m).evaluate(),
                       numpy.array([[0, 1, 2, 3],
                                    [1, 0, 1, 2],
                                    [2, 1, 0, 1]]))


def test_multiple_operations_with_structured_array():
    input = numpy.arange(12).reshape((4, 3))
    m0 = larray(input, shape=(4, 3))
    m1 = (m0 + 2) < 5
    m2 = (m0 < 5) + 2
    assert_array_equal(m1.evaluate(simplify=True), (input + 2) < 5)
    assert_array_equal(m2.evaluate(simplify=True), (input < 5) + 2)
    assert_array_equal(m0.evaluate(simplify=True), input)


def test_multiple_operations_with_functional_array():
    m = larray(lambda i: i, shape=(5,))
    m0 = m / 100.0
    m1 = 0.2 + m0
    assert_array_almost_equal(m0.evaluate(), numpy.array([0.0, 0.01, 0.02, 0.03, 0.04]), decimal=12)
    assert_array_almost_equal(m1.evaluate(), numpy.array([0.20, 0.21, 0.22, 0.23, 0.24]), decimal=12)
    assert_equal(m1[0], 0.2)


def test_operations_combining_constant_and_structured_arrays():
    m0 = larray(10, shape=(5,))
    m1 = larray(numpy.arange(5))
    m2 = m0 + m1
    assert_array_almost_equal(m2.evaluate(), numpy.arange(10, 15))


def test_apply_function_to_constant_array():
    f = lambda m: 2 * m + 3
    m0 = larray(5, shape=(4, 3))
    m1 = f(m0)
    assert isinstance(m1, larray)
    assert_equal(m1.evaluate(simplify=True), 13)
    # the following tests the internals, not the behaviour
    # it is just to check I understand what's going on
    assert_equal(m1.operations, [(operator.mul, 2), (operator.add, 3)])


def test_apply_function_to_structured_array():
    f = lambda m: 2 * m + 3
    input = numpy.arange(12).reshape((4, 3))
    m0 = larray(input, shape=(4, 3))
    m1 = f(m0)
    assert isinstance(m1, larray)
    assert_array_equal(m1.evaluate(simplify=True), input * 2 + 3)


def test_apply_function_to_functional_array():
    input = lambda i, j: 2 * i + j
    m0 = larray(input, shape=(4, 3))
    f = lambda m: 2 * m + 3
    m1 = f(m0)
    assert_array_equal(m1.evaluate(),
                       numpy.array([[3, 5, 7],
                                    [7, 9, 11],
                                    [11, 13, 15],
                                    [15, 17, 19]]))


def test_add_two_constant_arrays():
    m0 = larray(5, shape=(4, 3))
    m1 = larray(7, shape=(4, 3))
    m2 = m0 + m1
    assert_equal(m2.evaluate(simplify=True), 12)
    # the following tests the internals, not the behaviour
    # it is just to check I understand what's going on
    assert_equal(m2.base_value, m0.base_value)
    assert_equal(m2.operations, [(operator.add, m1)])


def test_add_incommensurate_arrays():
    m0 = larray(5, shape=(4, 3))
    m1 = larray(7, shape=(5, 3))
    assert_raises(ValueError, m0.__add__, m1)


def test_getitem_from_2D_constant_array():
    m = larray(3, shape=(4, 3))
    assert m[0, 0] == m[3, 2] == m[-1, 2] == m[-4, 2] == m[2, -3] == 3
    assert_raises(IndexError, m.__getitem__, (4, 0))
    assert_raises(IndexError, m.__getitem__, (2, -4))


def test_getitem_from_1D_constant_array():
    m = larray(3, shape=(43,))
    assert m[0] == m[42] == 3


def test_getitem__with_slice_from_constant_array():
    m = larray(3, shape=(4, 3))
    assert_array_equal(m[:3, 0],
                       numpy.array([3, 3, 3]))


def test_getitem__with_thinslice_from_constant_array():
    m = larray(3, shape=(4, 3))
    assert_equal(m[2:3, 0:1], 3)


def test_getitem__with_mask_from_constant_array():
    m = larray(3, shape=(4, 3))
    assert_array_equal(m[1, (0, 2)],
                       numpy.array([3, 3]))


def test_getitem_with_numpy_integers_from_2D_constant_array():
    if not hasattr(numpy, "int64"):
        raise SkipTest("test requires a 64-bit system")
    m = larray(3, shape=(4, 3))
    assert m[numpy.int64(0), numpy.int32(0)] == 3


def test_getslice_from_constant_array():
    m = larray(3, shape=(4, 3))
    assert_array_equal(m[:2],
                       numpy.array([[3, 3, 3],
                                    [3, 3, 3]]))


def test_getslice_past_bounds_from_constant_array():
    m = larray(3, shape=(5,))
    assert_array_equal(m[2:10],
                       numpy.array([3, 3, 3]))


def test_getitem_from_structured_array():
    m = larray(3 * numpy.ones((4, 3)), shape=(4, 3))
    assert m[0, 0] == m[3, 2] == m[-1, 2] == m[-4, 2] == m[2, -3] == 3
    assert_raises(IndexError, m.__getitem__, (4, 0))
    assert_raises(IndexError, m.__getitem__, (2, -4))


def test_getitem_from_2D_functional_array():
    m = larray(lambda i, j: 2 * i + j, shape=(6, 5))
    assert_equal(m[5, 4], 14)


def test_getitem_from_1D_functional_array():
    m = larray(lambda i: i ** 3, shape=(6,))
    assert_equal(m[5], 125)


def test_getitem_from_3D_functional_array():
    m = larray(lambda i, j, k: i + j + k, shape=(2, 3, 4))
    assert_raises(NotImplementedError, m.__getitem__, (0, 1, 2))


def test_getitem_from_vectorized_iterable():
    input = MockRNG(0, 1)
    m = larray(input, shape=(7,))
    m3 = m[3]
    assert isinstance(m3, (int, numpy.integer))
    assert_equal(m3, 0)
    assert_equal(m[0], 1)


def test_getitem_with_slice_from_2D_functional_array():
    m = larray(lambda i, j: 2 * i + j, shape=(6, 5))
    assert_array_equal(m[2:5, 3:],
                       numpy.array([[7, 8],
                                    [9, 10],
                                    [11, 12]]))


def test_getitem_with_slice_from_2D_functional_array_2():
    def test_function(i, j):
        return i * i + 2 * i * j + 3
    m = larray(test_function, shape=(3, 15))
    assert_array_equal(m[:, 3:14:3],
                       numpy.fromfunction(test_function, shape=(3, 15))[:, 3:14:3])


def test_getitem_with_mask_from_2D_functional_array():
    a = numpy.arange(30).reshape((6, 5))
    m = larray(lambda i, j: 5 * i + j, shape=(6, 5))
    assert_array_equal(a[[2, 3], [3, 4]],
                       numpy.array([13, 19]))
    assert_array_equal(m[[2, 3], [3, 4]],
                       numpy.array([13, 19]))


def test_getitem_with_mask_from_1D_functional_array():
    m = larray(lambda i: numpy.sqrt(i), shape=(10,))
    assert_array_equal(m[[0, 1, 4, 9]],
                       numpy.array([0, 1, 2, 3]))


def test_getitem_with_boolean_mask_from_1D_functional_array():
    m = larray(lambda i: numpy.sqrt(i), shape=(10,))
    assert_array_equal(m[numpy.array([1, 1, 0, 0, 1, 0, 0, 0, 0, 1], dtype=bool)],
                       numpy.array([0, 1, 2, 3]))


def test_getslice_from_2D_functional_array():
    m = larray(lambda i, j: 2 * i + j, shape=(6, 5))
    assert_array_equal(m[1:3],
                       numpy.array([[2, 3, 4, 5, 6],
                                    [4, 5, 6, 7, 8]]))


def test_getitem_from_iterator_array():
    m = larray(iter([1, 2, 3]), shape=(3,))
    assert_raises(NotImplementedError, m.__getitem__, 2)


def test_getitem_from_array_with_operations():
    a1 = numpy.array([[1, 3, 5], [7, 9, 11]])
    m1 = larray(a1)
    f = lambda i, j: numpy.sqrt(i * i + j * j)
    a2 = numpy.fromfunction(f, shape=(2, 3))
    m2 = larray(f, shape=(2, 3))
    a3 = 3 * a1 + a2
    m3 = 3 * m1 + m2
    assert_array_equal(a3[:, (0, 2)],
                       m3[:, (0, 2)])


def test_evaluate_with_invalid_base_value():
    m = larray(range(5))
    m.base_value = "foo"
    assert_raises(ValueError, m.evaluate)


def test_partially_evaluate_with_invalid_base_value():
    m = larray(range(5))
    m.base_value = "foo"
    assert_raises(ValueError, m._partially_evaluate, 3)


def test_check_bounds_with_invalid_address():
    m = larray([[1, 3, 5], [7, 9, 11]])
    assert_raises(TypeError, m.check_bounds, (object(), 1))


def test_check_bounds_with_invalid_address2():
    m = larray([[1, 3, 5], [7, 9, 11]])
    assert_raises(ValueError, m.check_bounds, ([], 1))


def test_partially_evaluate_constant_array_with_one_element():
    m = larray(3, shape=(1,))
    a = 3 * numpy.ones((1,))
    m1 = larray(3, shape=(1, 1))
    a1 = 3 * numpy.ones((1, 1))
    m2 = larray(3, shape=(1, 1, 1))
    a2 = 3 * numpy.ones((1, 1, 1))
    assert_equal(a[0], m[0])
    assert_equal(a.shape, m.shape)
    assert_equal(a[:].shape, m[:].shape)
    assert_equal(a, m.evaluate())
    assert_equal(a1.shape, m1.shape)
    assert_equal(a1[0,:].shape, m1[0,:].shape)
    assert_equal(a1[:].shape, m1[:].shape)
    assert_equal(a1, m1.evaluate())
    assert_equal(a2.shape, m2.shape)
    assert_equal(a2[:, 0,:].shape, m2[:, 0,:].shape)
    assert_equal(a2[:].shape, m2[:].shape)
    assert_equal(a2, m2.evaluate())


def test_partially_evaluate_constant_array_with_boolean_index():
    m = larray(3, shape=(4, 5))
    a = 3 * numpy.ones((4, 5))
    addr_bool = numpy.array([True, True, False, False, True])
    addr_int = numpy.array([0, 1, 4])
    assert_equal(a[::2, addr_bool].shape, a[::2, addr_int].shape)
    assert_equal(a[::2, addr_int].shape, m[::2, addr_int].shape)
    assert_equal(a[::2, addr_bool].shape, m[::2, addr_bool].shape)


def test_partially_evaluate_constant_array_with_all_boolean_indices_false():
    m = larray(3, shape=(3,))
    a = 3 * numpy.ones((3,))
    addr_bool = numpy.array([False, False, False])
    assert_equal(a[addr_bool].shape, m[addr_bool].shape)


def test_partially_evaluate_constant_array_with_only_one_boolean_indice_true():
    m = larray(3, shape=(3,))
    a = 3 * numpy.ones((3,))
    addr_bool = numpy.array([False, True, False])
    assert_equal(a[addr_bool].shape, m[addr_bool].shape)
    assert_equal(m[addr_bool][0], a[0])


def test_partially_evaluate_constant_array_with_boolean_indice_as_random_valid_ndarray():
    m = larray(3, shape=(3,))
    a = 3 * numpy.ones((3,))
    addr_bool = numpy.random.rand(3) > 0.5
    while not addr_bool.any():
        # random array, but not [False, False, False]
        addr_bool = numpy.random.rand(3) > 0.5
    assert_equal(a[addr_bool].shape, m[addr_bool].shape)
    assert_equal(m[addr_bool][0], a[addr_bool][0])


def test_partially_evaluate_constant_array_size_one_with_boolean_index_true():
    m = larray(3, shape=(1,))
    a = numpy.array([3])
    addr_bool = numpy.array([True])
    m1 = larray(3, shape=(1, 1))
    a1 = 3 * numpy.ones((1, 1))
    addr_bool1 = numpy.array([[True]], ndmin=2)
    assert_equal(m[addr_bool][0], a[0])
    assert_equal(m[addr_bool], a[addr_bool])
    assert_equal(m[addr_bool].shape, a[addr_bool].shape)
    assert_equal(m1[addr_bool1][0], a1[addr_bool1][0])
    assert_equal(m1[addr_bool1].shape, a1[addr_bool1].shape)


def test_partially_evaluate_constant_array_size_two_with_boolean_index_true():
    m2 = larray(3, shape=(1, 2))
    a2 = 3 * numpy.ones((1, 2))
    addr_bool2 = numpy.ones((1, 2), dtype=bool)
    assert_equal(m2[addr_bool2][0], a2[addr_bool2][0])
    assert_equal(m2[addr_bool2].shape, a2[addr_bool2].shape)


def test_partially_evaluate_constant_array_size_one_with_boolean_index_false():
    m = larray(3, shape=(1,))
    m1 = larray(3, shape=(1, 1))
    a = numpy.array([3])
    a1 = numpy.array([[3]], ndmin=2)
    addr_bool = numpy.array([False])
    addr_bool1 = numpy.array([[False]], ndmin=2)
    addr_bool2 = numpy.array([False])
    assert_equal(m[addr_bool].shape, a[addr_bool].shape)
    assert_equal(m1[addr_bool1].shape, a1[addr_bool1].shape)


def test_partially_evaluate_constant_array_size_with_empty_boolean_index():
    m = larray(3, shape=(1,))
    a = numpy.array([3])
    addr_bool = numpy.array([], dtype='bool')
    assert_equal(m[addr_bool].shape, a[addr_bool].shape)
    assert_equal(m[addr_bool].shape, (0,))


def test_partially_evaluate_functional_array_with_boolean_index():
    m = larray(lambda i, j: 5 * i + j, shape=(4, 5))
    a = numpy.arange(20.0).reshape((4, 5))
    addr_bool = numpy.array([True, True, False, False, True])
    addr_int = numpy.array([0, 1, 4])
    assert_equal(a[::2, addr_bool].shape, a[::2, addr_int].shape)
    assert_equal(a[::2, addr_int].shape, m[::2, addr_int].shape)
    assert_equal(a[::2, addr_bool].shape, m[::2, addr_bool].shape)


def test_getslice_with_vectorized_iterable():
    input = MockRNG(0, 1)
    m = larray(input, shape=(7, 3))
    assert_array_equal(m[::2, (0, 2)],
                       numpy.arange(8).reshape((4, 2)))


def test_equality():
    m1 = larray(42.0, shape=(4, 5)) / 23.0 + 2.0
    m2 = larray(42.0, shape=(4, 5)) / 23.0 + 2.0
    assert_equal(m1, m2)


def test_deepcopy():
    m1 = 3 * larray(lambda i, j: 5 * i + j, shape=(4, 5)) + 2
    m2 = deepcopy(m1)
    m1.shape = (3, 4)
    m3 = deepcopy(m1)
    assert_equal(m1.shape, m3.shape, (3, 4))
    assert_equal(m2.shape, (4, 5))
    assert_array_equal(m1.evaluate(), m3.evaluate())


def test_deepcopy_with_ufunc():
    m1 = sqrt(larray([x ** 2 for x in range(5)]))
    m2 = deepcopy(m1)
    m1.base_value[0] = 49
    assert_array_equal(m1.evaluate(), numpy.array([7, 1, 2, 3, 4]))
    assert_array_equal(m2.evaluate(), numpy.array([0, 1, 2, 3, 4]))


def test_set_shape():
    m = larray(42) + larray(lambda i: 3 * i)
    assert_equal(m.shape, None)
    m.shape = (5,)
    assert_array_equal(m.evaluate(), numpy.array([42, 45, 48, 51, 54]))


def test_call():
    A = larray(numpy.array([1, 2, 3]), shape=(3,)) - 1
    B = 0.5 * larray(lambda i: 2 * i, shape=(3,))
    C = B(A)
    assert_array_equal(C.evaluate(), numpy.array([0, 1, 2]))
    assert_array_equal(A.evaluate(), numpy.array([0, 1, 2]))  # A should be unchanged


def test_call2():
    positions = numpy.array(
        [[0.,  2.,  4.,  6.,  8.],
         [0.,  0.,  0.,  0.,  0.],
         [0.,  0.,  0.,  0.,  0.]])

    def position_generator(i):
        return positions.T[i]

    def distances(A, B):
        d = A - B
        d **= 2
        d = numpy.sum(d, axis=-1)
        numpy.sqrt(d, d)
        return d

    def distance_generator(f, g):
        def distance_map(i, j):
            return distances(f(i), g(j))
        return distance_map
    distance_map = larray(distance_generator(position_generator, position_generator),
                          shape=(4, 5))
    f_delay = 1000 * larray(lambda d: 0.1 * (1 + d), shape=(4, 5))
    assert_array_almost_equal(
        f_delay(distance_map).evaluate(),
        numpy.array([[100, 300, 500, 700, 900],
                     [300, 100, 300, 500, 700],
                     [500, 300, 100, 300, 500],
                     [700, 500, 300, 100, 300]], dtype=float),
        decimal=12)
    # repeat, should be idempotent
    assert_array_almost_equal(
        f_delay(distance_map).evaluate(),
        numpy.array([[100, 300, 500, 700, 900],
                     [300, 100, 300, 500, 700],
                     [500, 300, 100, 300, 500],
                     [700, 500, 300, 100, 300]], dtype=float),
        decimal=12)


def test__issue4():
    # In order to avoid the errors associated with version changes of numpy, mask1 and mask2 no longer contain boolean values ​​but integer values
    a = numpy.arange(12).reshape((4, 3))
    b = larray(numpy.arange(12).reshape((4, 3)))
    mask1 = (slice(None), int(True))
    mask2 = (slice(None), numpy.array([int(True)]))
    assert_equal(b[mask1].shape, partial_shape(mask1, b.shape), a[mask1].shape)
    assert_equal(b[mask2].shape, partial_shape(mask2, b.shape), a[mask2].shape)


def test__issue3():
    a = numpy.arange(12).reshape((4, 3))
    b = larray(a)
    c = larray(lambda i, j: 3*i + j, shape=(4, 3))
    assert_array_equal(a[(1, 3), :][:, (0, 2)], b[(1, 3), :][:, (0, 2)])
    assert_array_equal(b[(1, 3), :][:, (0, 2)], c[(1, 3), :][:, (0, 2)])
    assert_array_equal(a[(1, 3), (0, 2)], b[(1, 3), (0, 2)])
    assert_array_equal(b[(1, 3), (0, 2)], c[(1, 3), (0, 2)])


def test_partial_shape():
    a = numpy.arange(12).reshape((4, 3))
    test_cases = [
        (slice(None), (4, 3)),
        ((slice(None), slice(None)), (4, 3)),
        (slice(1, None, 2), (2, 3)),
        (1, (3,)),
        ((1, slice(None)), (3,)),
        ([0, 2, 3], (3, 3)),
        (numpy.array([0, 2, 3]), (3, 3)),
        ((numpy.array([0, 2, 3]), slice(None)), (3, 3)),
        (numpy.array([True, False, True, True]), (3, 3)),
        #(numpy.array([True, False]), (1, 3)),  # not valid with NumPy 1.13
        (numpy.array([[True, False, False], [False, False, False], [True, True, False], [False, True, False]]), (4,)),
        #(numpy.array([[True, False, False], [False, False, False], [True, True, False]]), (3,)),  # not valid with NumPy 1.13
        ((3, 1), tuple()),
        ((slice(None), 1), (4,)),
        ((slice(None), slice(1, None, 3)), (4, 1)),
        ((numpy.array([0, 3]), 2), (2,)),
        ((numpy.array([0, 3]), numpy.array([1, 2])), (2,)),
        ((slice(None), numpy.array([2])), (4, 1)),
        (((1, 3), (0, 2)), (2,)),
        (numpy.array([], bool), (0, 3)),
    ]
    for mask, expected_shape in test_cases:
        assert_equal(partial_shape(mask, a.shape), a[mask].shape)
        assert_equal(partial_shape(mask, a.shape), expected_shape)
    b = numpy.arange(5)
    test_cases = [
        (numpy.arange(5), (5,))
    ]
    for mask, expected_shape in test_cases:
        assert_equal(partial_shape(mask, b.shape), b[mask].shape)
        assert_equal(partial_shape(mask, b.shape), expected_shape)

def test_is_homogeneous():
    m0 = larray(10, shape=(5,))
    m1 = larray(numpy.arange(1, 6))
    m2 = m0 + m1
    m3 = 9 + m0 / m1
    assert m0.is_homogeneous
    assert not m1.is_homogeneous
    assert not m2.is_homogeneous
    assert not m3.is_homogeneous
